/*
			 Salvatore Ruggieri (c), 2002-2005
*/

/** @file YaDT.h
  *
  * This is the header of the YaDT project.
  */ 

#if !defined(_YADT_H__INCLUDED_)
#define _YADT_H__INCLUDED_

#include <string>
#include <vector>
#include <stdexcept>
#include <iostream> /* standard iostream and fstream */
#include <fstream> 

#ifndef DOXYGEN_SHOULD_SKIP_THIS
using namespace std;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \mainpage YaDT: Yet another Decision Tree builder
 *
 * YaDT is a new from-scratch implementation of the 
 * entropy-based tree construction algorithm. It has been 
 * designed and implemented in C++ with strong emphasis 
 * on efficiency (time and space) and portability on Windows and Linux.
 * <p>
 * <center>\link yadt YaDT namespace \endlink</center><br>
 * <p>
 */

/** The Yet another Decision Tree builder namespace. */
namespace yadt
{
	/** YaDT version string. */
	const char *get_version();
	/** YaDT build number. */
	const int get_build();
	/** YaDT copyright string. */
	const char *get_copyright();


	/** Source of textual data. Data sources include: text files, gzipped text files,
		JDBC relational tables. The purpose of this class is to encapsulate information
		on accessing data. It is used in methods that need to read data. 
			@see table::load_meta_data()
			@see table::load_data()
			@see dtree::predict()
			@see dtree::evaluate()
	*/
	class datasource
	{
	public:
		/** Constructor.
		  * Takes a data source specification: "FILE\#filename\#c" specifies a file with
		  * name "filename" and where column separator is the char 'c'; 
		  * "FILE\#filename.gz\#c" specifies a gzipped file with name "filename.gz" and 
		  * where column separator is the char 'c';
		  * "JDBC\#driver\#url\#usr\#pwd\#sql" -- to be added.
		  *   @param %specs data source specification.
		  */
		datasource(const string &specs);
		/** Destructor. */
		~datasource();
	private:
	#ifndef DOXYGEN_SHOULD_SKIP_THIS
			friend class dtree;
			friend class table;
	#endif /* DOXYGEN_SHOULD_SKIP_THIS */
		string specifications;
	};

	// forward references
	class conf_matrix;
	class dtree;

	/** A relational table. A relational table is a collection of attributes
	  * (discrete, continuous), of a class and, possibly, of a weights column. 
	  */
	class table
	{
	public:
		/** Constructor.
		  *   @param name %table name.
		  */
		table(const string &name);
		/** Destructor. */
		~table();
		/** Copy constructor not defined. */
		table( const table& );
		/** Assignment constructor not defined. */
		const table& operator=( const table& );
		/** Load metadata from a datasource. */
		void load_meta_data(const datasource &ds)
				 throw(runtime_error);
		/** Return table name. */
		string get_name() const;

		/** Load a table from a datasource. 
		  *    @param ds data source
		  *    @param unknown the unknown token
		  */
		void load_data(const datasource &ds, const string &unknown = "?") 
			throw(runtime_error);
		/** Return number of columns in the table. */
		unsigned get_no_columns() const;
		/** Return number of rows in the %table. */
		unsigned get_no_rows() const;

		/** Binary output. Attention: binary input/output is not guarranteed
		  * to be consistent among different releases of this library!!
		  *   @param filename the input filename. */
		void toBinary(const string &filename) const; 
		/** Binary input. Attention: binary input/output is not guarranteed
		  * to be consistent among different releases of this library!!
		  *   @param filename the output filename. */
		static table* fromBinary(const string &filename); 

		/* utility methods */

		/** Set output log. Default value is NULL, i.e. no output.
		  *     @param new_log output log.
		  *     @return old output log.	  */
		ostream *set_log(ostream *new_log);
		/** Set output log verbosity of operations.
		  * Verbosity values are: 0 = none, 1 low, 2 normal, 3 high.
		  *     @param newverbosity new verbosity of output.
		  *     @return old verbosity of output. */
		unsigned set_verbosity(unsigned newverbosity);
		/** Return elapsed time for lastloading of table. In other words, this
		  * method returns the elapsed time of the last call to the 
		  * load_data() method. */
		double get_elapsed() const;
		/** XML output of data dictionary. PMML 2.0 complaint. */
		void toXML_data_dictionary(ostream &os = cout) const; 
		/** XML output of mining schema.  PMML 2.0 complaint. */
		void toXML_mining_schema(ostream &os = cout) const; 

		/** A subset of a table. The purpose of this class is to
		  * encapsulate information on accessing a subset of a table. 
		  * A subset can only be obtained
		  * as the return value of some computation.
		  * @see dtree::build()
		  * @see dtree::predict() 
		  * @see table::get_wsubset_all() 
		  */
		class subset
		{
		public:
			/** Destructor. */
			~subset();
			/** Copy constructor not defined. */
			subset( const subset& );
			/** Assignment constructor not defined. */
			const subset& operator=( const subset& );
		private:
	#ifndef DOXYGEN_SHOULD_SKIP_THIS
			friend class dtree;
			friend class table;
	#endif /* DOXYGEN_SHOULD_SKIP_THIS */
			/** Private constructor. */
			subset(void *actual);
			void *real;
		};

		/** Return new weighted subset for all table rows. Weights of cases (i.e., rows) in the
		  * returned subset are assigned accordingly to the weights column, if
		  * present or to the default value (1.0) otherwise. */
		subset* get_wsubset_all() const;
		/** Return new weighted subset for first n table rows. Weights of cases (i.e., rows) in the
		  * returned subset are assigned accordingly to the weights column, if
		  * present or to the default value (1.0) otherwise. */
		subset* get_wsubset_first_n(unsigned n) const;
		/** Return new weighted subset for n randomly selected (no repetition) table rows. 
		  * Weights of cases (i.e., rows) in the
		  * returned w_sequence are assigned accordingly to the weights column, if
		  * present or to the default value (1.0) otherwise. */
		subset* get_wsubset_random(unsigned n) const;
		/** Return new weighted subset as difference from a given one. 
		  * The returned subset consists of all tables rows not present in the passed
		  * subset. Weights of cases (i.e., rows) in the
		  * returned w_sequence are assigned accordingly to the weights column, if
		  * present or to the default value (1.0) otherwise. */
		subset* get_wsubset_difference(subset *subtable) const;

	private:
	#ifndef DOXYGEN_SHOULD_SKIP_THIS
		friend class conf_matrix;
		friend class dtree;
	#endif /* DOXYGEN_SHOULD_SKIP_THIS */
		/** Actual table class. */
		void *real;
	};

	/**
	  * Models a confusion matrix. A confusion matrix is a matrix n x n 
	  * of weight_type::type, where n is the number of distinct values of a class
	  * (including the "unknown" class value). 
	  * For a %conf_matrix object cm, cm.get_element(act, pred) represents the weighted sum of 
	  * cases of class with index act predicted as belonging to class of index pred.
	  * The "unknown" class value has index 0: cm.get_element(0, x) should be 0 since actual 
	  * classes should never be "unknown" (this may happen, however, if the actual
	  * class of cases in the test set did not appear in the training set); 
	  * cm.get_element(x, 0) should be 0 since predicted classes should never be "unknown".
	  * A confusion matrix can only be obtained
	  * as the return value of some computation.
	  * @see dtree::predict
	  */
	class conf_matrix 
	{
	public:
		/** Destructor. */
		~conf_matrix();
		/** Copy constructor not defined. */
		conf_matrix( const conf_matrix& );
		/** Assignment constructor not defined. */
		const conf_matrix& operator=( const conf_matrix& );
		/** Clone method. Return a newly allocated %conf_matrix that is a clone
		  * of the called object.
		  */
		conf_matrix* clone() const;
		/** Return degree of the matrix. The degree of a matrix n x n, is n. */
		size_t size() const;
		/** Return a cell of the matrix. */
		float get_element(size_t actual, size_t predicted) const;
		/** Return misclassification percentage of the confusion matrix.
		  * The misclassification percentage of a %conf_matrix object cm
		  * is sum(i != j, cm[i, j]).
		  */
		float mis_perc() const;
		/** Return total weights in the confusion matrix. More formally,
		  * for a %conf_matrix object cm, it is returned sum(true, cm[i, j]).
		  */
		float cases() const;
		/** Return elapsed time taken to build the confusion matrix. */
		double get_elapsed() const;
		/** Text output of the confusion matrix.
		  *     @param os output stream.
		  *     @param space indent space.
		  */
		void toTEXT(ostream& os = cout, unsigned space = 0) const;
		/** XML output of the confusion matrix. */
		void toXML(ostream& os = cout) const;

	private:
	#ifndef DOXYGEN_SHOULD_SKIP_THIS
		friend class dtree;
	#endif /* DOXYGEN_SHOULD_SKIP_THIS */
		/** Private constructor. A confusion matrix can only be obtained
		  * as the return value of some computation.
		  * @see dtree::predict
		  */
		conf_matrix(table *maintable)
			throw(runtime_error);
		conf_matrix(const conf_matrix *cm)
			throw(runtime_error);
		/** Actual class. */
		void *real;
	};

	/**
	  * A decision tree. The class provides methods for building, simplifying,
	  * and evaluating a decision tree.
	  */
	class dtree 
	{
	public:
		/** Constructor. @param name decision tree name. */
		dtree(const string &name="my_decision_tree");
		/** Destructor. */
		~dtree();
		/** Copy constructor not defined. */
		dtree( const dtree& );
		/** Assignment constructor not defined. */
		const dtree& operator=( const dtree& );
		/** Return a clone of the called object. */
		dtree* clone() const;

		/** Return elapsed time (in secs) taken to build tree. */
		double get_elapsed() const;
		/** Return number of tree nodes. */
		unsigned size() const;
		/** Return tree depth. */
		unsigned depth() const;
		/** Return number of cases used in building the tree. */
		unsigned training_n_rows() const;

		/** Pruning strategy options. */
		typedef enum {
			/** No pruning. */
			PRUNING_NO, 
			/** C4.5 pruning strategy. */
			PRUNING_C45, 
			/** YaDT pruning strategy. */
			PRUNING_DT
		} PruningStrategy;

		/** Split type of a splitting decision node. */
		typedef enum {
			/** Information gain split. */
			ST_GAIN,
			/** Information gain ratio split. */
			ST_GAIN_RATIO
		} SplitType;

		/** Set simplification strategy.
		  *   @param strategy new pruning strategy.
		  *   @return true if setting the new strategy succeeded. */
		bool set_pruning_strategy(PruningStrategy strategy);
		/** Set split strategy.
		  *   @param st new split strategy.
		  *   @return true if setting the new strategy succeeded. */
		bool set_split_type(SplitType st);
		/** Set mininum weight of cases in sons in order to further split a node
		  *  during tree building. Default value is 2.0. Any value must be > 0.
		  *   @param min_objects new minimum weight.
		  *   @return true if setting the new weight succeeded. */
		bool set_min_obj(float min_objects);
		/** Set confidence level in simplifying a decision tree. The new
		  * confidence level must be in the range [0,1].*/
		bool set_conf_level(float conf_level);

		/**  Build an pruned tree. The method builds a tree and simplifies it.
		  *    @param maintable a table containing the training set.
		  *    @param subtable the subset of maintable used as training set. 
		  *            NULL value denotes all the table as training set.
		  *    @param evaluate true if a confusion matrix must be also computed. The
		  *            resulting confusion matrix can be obtained by calling the
		  *            get_prediction() method.
		  *    @see set_pruning_strategy(), set_split_type(), set_min_obj(), set_conf_level()
		  */
		void build(table* maintable, table::subset *subtable, bool evaluate = true)
			throw (runtime_error);
		/** Return confusion matrix over the training set. The method returns NULL if no
		  * tree was build or it was build by not requiring the computation of a confusion
		  * matrix. 
		  * @see build_unpruned(), build_pruned()
		  */
		conf_matrix* get_prediction();
		/** Test classes of unseen cases. The source rst::inpout::stream is required to 
		  * provide for each case all attributes in the same order as the columns
		  * of training set and then the actual class: no weights must be provided.
		  *    @param ds provider of unseen cases.
		  *    @return a confusion matrix comparing prediction against actual classes.
		  */
		conf_matrix* predict(const datasource &ds) const
			throw(runtime_error);
		/** Predict classes of unseen cases. The source rst::inpout::stream is required to 
		  * provide for each case all attributes in the same order as the columns
		  * of training set: no class or weights must be provided. Optionally, a 
		  * further attribute may be provided (tipically a key of the case) that is
		  * produced in output together with predicted class and confidence.
		  *    @param ds provider of unseen cases.
		  *    @param output output stream of predictions.
		  *    @param sep %column separator in output stream.
		  *	   @returns elapsed time.
		  */
		double evaluate(const datasource &ds, ostream &output, char sep = '\t') const
			throw(runtime_error);
		/** Predict class and confidence of an unseen case. The attributes of the case are 
		  * provided as a vector of C strings in the same order as the columns
		  * of training set: no class must be provided. Optionally, a case weight may be
		  * provided (which affects confidence of prediction).
		  *    @param attributes vector of C string representing case attributes.
		  *    @param weight case weight.
		  *    @return a pair with predicted class and confidence.
		  */
		pair<string, float> predict(vector<string> &attributes, float weight = 1) const;
		// predict all cases in a given table
		/** Test classes of unseen cases. The cases table is required to 
		  * provide for each case all attributes in the same order as the columns
		  * of training set and the actual class of cases. If a weights %column is
		  * prese: no weights must be provided.
		  *    @param cases a table containing the unseen cases.
		  *    @param subtable the subset of cases to test. 
		  *            NULL value denotes all the table.
		  *    @return a confusion matrix comparing prediction against actual classes.
		  */
		conf_matrix* predict(table* cases, table::subset *subtable) const
			throw(runtime_error);

		/** Textual output. */
		void toTEXT(ostream& os = cout) const;
		/** Dot output. */
		void toDOT(ostream& os = cout) const;
		/** XML output. PMML 2.0 complaint. */
		void toXML(ostream &os = cout, const conf_matrix *cmTest = NULL) const; 

		/** Binary output. 
		  *   @param filename the output filename. */
		void toBinary(const string &filename);
		/** Binary input. 
		  *   @param filename the input filename.
		  *   @returns a newly allocated decision tree. */
		static dtree *fromBinary(const string &filename);

	private:
		/** Actual decision tree. */
		void *real;
	};

} // namespace yadt

using namespace yadt;

#endif // !defined(_YADT_H__INCLUDED_)
